import git.exc
from smartsquash.helpers import run_rebase
from typing import Dict, Set, Optional


def get_files_changed_in_staging(repo: git.Repo) -> Set[str]:
    output: Set[str] = (repo.git.diff("--name-only", "--cached", "-r").splitlines())
    return set(output)


def get_closest_change_commit(
    files_changed: Set[str], commit_changed_files: Dict[str, Set[str]]
) -> Optional[str]:
    for commit_sha, files in commit_changed_files.items():
        if files_changed.issubset(files):
            return commit_sha
    return None


def run_fixup(
    repo: git.Repo, target_branch: str, fixup_commit_sha: str, add: bool, dry: bool
):
    command = ["--fixup", fixup_commit_sha]
    if add:
        command.insert(0, "-a")
    repo.git.commit(*command)
    run_rebase(repo, target_branch, "true", dry, autosquash=True)



import git
import git.exc
import itertools
import subprocess
import collections
from typing import List, Tuple, Dict, Set
from smartsquash import helpers
from loguru import logger


def get_commits_in_between(
    commit_a: git.Commit, commit_b: git.Commit, commits: List[git.Commit]
) -> List[git.Commit]:
    """Returns commits from """
    find_a_index = -1
    find_b_index = -1
    for index, commit in enumerate(commits):
        if commit_a.hexsha == commit.hexsha:
            find_a_index = index
        if commit_b.hexsha == commit.hexsha:
            find_b_index = index
        if find_a_index != -1 and find_b_index != -1:
            break
    first: int = find_a_index if find_a_index < find_b_index else find_b_index
    second: int = find_b_index if first == find_a_index else find_a_index
    return commits[first + 1 : second]


def has_relevant_files_changed_in_between(
    commit_a: git.Commit,
    commit_b: git.Commit,
    commits: List[git.Commit],
    commit_changed_files: Dict[str, Set[str]],
) -> bool:
    """
    we should look for other files that changed in between commits AND
    in addition also change 'our files'(the files from 'commit_a,commit_b')

    We don't care about other files that may have changed in between
    """
    commits_in_between = get_commits_in_between(commit_a, commit_b, commits)
    files = commit_changed_files[commit_a.hexsha]
    for commit in commits_in_between:
        for file in commit_changed_files[commit.hexsha]:
            if file in files and commit_changed_files[commit.hexsha] != files:
                return True
    return False


def commit_diff_empty(commit_a: git.Commit, commit_b: git.Commit) -> bool:
    output = subprocess.check_output(
        ["git", "diff", f"{commit_a.hexsha}...{commit_b.hexsha}"],
        cwd=commit_a.repo.working_dir,
    ).decode()
    return not output


def has_same_files(files_a: List[str], files_b: List[str]) -> bool:
    files_a: List[str] = sorted(files_a)
    files_b: List[str] = sorted(files_b)
    return files_a == files_b


def get_commit_message(commit: git.Commit) -> str:
    if commit.message:
        return commit.message.split("\n")[0]
    return ""


def fulfils_conditions(
    commit_a: git.Commit,
    commit_b: git.Commit,
    commits: List[git.Commit],
    commit_changed_files: Dict[str, Set[str]],
) -> bool:
    if not has_same_files(
        list(commit_changed_files[commit_a.hexsha]),
        list(commit_changed_files[commit_b.hexsha]),
    ):
        return False

    if has_relevant_files_changed_in_between(
        commit_a, commit_b, commits, commit_changed_files
    ):
        return False
    if commit_diff_empty(commit_a, commit_b):
        return False
    return True


def get_squash_combinations(
    commits: List[git.Commit], commit_changed_files: Dict[str, Set[str]]
):
    squash_combinations: Dict[str, List[git.Commit]] = collections.defaultdict(list)
    for commit_a, commit_b in itertools.combinations(commits, 2):
        if not fulfils_conditions(commit_a, commit_b, commits, commit_changed_files):
            continue
        squash_combinations[commit_a.hexsha].append(commit_b)
    return squash_combinations


def get_rebase_data(commits: List[git.Commit]) -> Tuple[bool, str]:
    has_rebase = False
    data = ""
    logger.info("Fetching files changed by commits...")
    commit_changed_files: Dict[str, Set[str]] = helpers.get_commits_changed_files(
        commits
    )
    logger.info(f"Comparing {len(commits)} commits with each other...")
    squash_combinations: Dict[str, List[git.Commit]] = get_squash_combinations(
        commits, commit_changed_files
    )

    squash_list: List[str] = []
    for commit in commits:
        if commit.hexsha in squash_list:
            continue
        if to_squash_commits := squash_combinations.get(commit.hexsha):
            # commit has commits to be squashed
            data += f"pick {commit.hexsha[:7]} {get_commit_message(commit)}\n"
            for to_squash in to_squash_commits:
                has_rebase = True
                data += (
                    f"fixup {to_squash.hexsha[:7]} {get_commit_message(to_squash)}\n"
                )
                squash_list.append(to_squash.hexsha)
        else:
            data += f"pick {commit.hexsha[:7]} {get_commit_message(commit)}\n"
    return has_rebase, data



import git
import sys
from typing import List, Dict, Set, Optional
from smartsquash.helpers import (
    retrieve_commits,
    get_commits_changed_files,
    run_rebase,
)
from smartsquash.fixup import (
    run_fixup,
    get_closest_change_commit,
    get_files_changed_in_staging,
)
from smartsquash.squash import get_rebase_data
from loguru import logger


def squash(target_branch: str, repo: git.Repo, dry: bool = False):
    all_commits: List[git.Commit] = retrieve_commits(repo, target_branch)
    has_rebase, rebase_data = get_rebase_data(all_commits)
    rebase_data: str = rebase_data.replace("'", "", -1)
    if has_rebase:
        run_rebase(
            repo, target_branch, f"echo '{rebase_data}' >", dry, autosquash=False
        )


def fixup(
    target_branch: str, repo: git.Repo, add: bool = False, dry: bool = False
) -> bool:
    if not repo.is_dirty():
        logger.info("Repository is not dirty. No files to fixup.")
        return False
    commits: List[git.Commit] = retrieve_commits(repo, target_branch, reverse=False)
    logger.info("Fetching files changed by commits...")
    commit_changed_files: Dict[str, Set[str]] = get_commits_changed_files(commits)
    files_changed: Set[str] = get_files_changed_in_staging(repo)
    closes_change_commit: Optional[str] = get_closest_change_commit(
        files_changed, commit_changed_files
    )
    if closes_change_commit:
        if dry:
            sys.exit(0)
        run_fixup(repo, target_branch, closes_change_commit, add, dry)
        return True
    logger.error("No commits found to fixup. You'll need to fixup manually")
    return False



import git
import git.exc
import os
from pathlib import Path
import subprocess
import sys
import collections
import enum
from typing import List, Dict, Set, Optional
from smartsquash.decorators import memorize_files_changed
from loguru import logger


class ErrorMessages(enum.Enum):
    HEAD_DETACHED = "HEAD is detached. Exiting"
    TARGET_EQUALS_CURRENT = "Target branch equals current active branch. Exiting"
    NOT_A_GIT_REPO = "The target is not a git repository. Exiting"
    PATH_NOT_EXIST = "The path doesn't exist. Exiting"

    def __str__(self):
        return self.value


def get_repo(repo_path: str, target_branch: str) -> git.Repo:
    if not Path(repo_path).exists():
        sys.exit(ErrorMessages.PATH_NOT_EXIST)
    try:
        repo: git.Repo = git.Repo(repo_path, search_parent_directories=True)
    except git.exc.InvalidGitRepositoryError:
        sys.exit(str(ErrorMessages.NOT_A_GIT_REPO))
    if message := get_repo_invalid_message(repo, target_branch):
        sys.exit(message)
    return repo


def retrieve_commits(
    repo: git.Repo, target_branch: str, reverse: bool = True
) -> List[git.Commit]:
    """
    retrieves commits that are only part of the currently active branch,
    and are not in the target branch
    - git cherry could be used for this, but GitPython doesn't support it
    - Just run raw git command, if this becomes bottleneck
    """
    target_commits_sha: List[git.Commit] = [
        commit.hexsha
        for commit in repo.iter_commits(rev=target_branch)
        if len(commit.parents) < 2  # ignore merge commits
    ]
    commits: List[git.Commit] = [
        commit
        for commit in repo.iter_commits(rev=repo.active_branch)
        if commit.hexsha not in target_commits_sha
        and len(commit.parents) < 2  # ignore merge commits
    ]
    if reverse:
        commits.reverse()
    return commits


@memorize_files_changed
def files_changed_by_commit(working_dir: str, commit: str) -> List[str]:
    output = (
        subprocess.check_output(
            ["git", "diff-tree", "--no-commit-id", "--name-only", "-r", commit],
            cwd=working_dir,
        )
        .decode()
        .splitlines()
    )
    return [output for output in output if output]


def get_commits_changed_files(commits: List[git.Commit]) -> Dict[str, Set[str]]:
    commit_changed: Dict[str, Set[str]] = collections.defaultdict(set)
    for commit in commits:
        for file in files_changed_by_commit(commit.repo.working_dir, commit.hexsha):
            commit_changed[commit.hexsha].add(file)
    return commit_changed


def get_repo_invalid_message(repo: git.Repo, target_branch: str) -> Optional[str]:
    if repo.head.is_detached:
        return str(ErrorMessages.HEAD_DETACHED)
    if repo.active_branch.name == target_branch:
        return str(ErrorMessages.TARGET_EQUALS_CURRENT)
    return None


def run_rebase(
    repo: git.Repo,
    target_branch: str,
    sequence_editor: str,
    dry: bool = False,
    autosquash: bool = True,
):
    os.environ["GIT_SEQUENCE_EDITOR"] = sequence_editor
    args: List[str] = ["-i", target_branch]
    if dry:
        sys.exit(0)
    if autosquash:
        args.insert(0, "--autosquash")
    try:
        repo.git.rebase(args)
        print("Rebase done")
    except git.CommandError:
        repo.git.rebase("--abort")
        logger.error("Rebase failed and aborted. You'll need to squash manually")



import functools
from typing import Callable, Dict, List


def memorize_files_changed(func) -> Callable:
    @functools.wraps(func)
    def wrapper_memorize_paths(*args, **kwargs) -> Dict[str, List[str]]:
        commit_sha: str = args[1]

        if not hasattr(wrapper_memorize_paths, "paths"):
            wrapper_memorize_paths.paths = {}
        if not wrapper_memorize_paths.paths.get(commit_sha):
            wrapper_memorize_paths.paths[commit_sha] = func(*args, **kwargs)
        return wrapper_memorize_paths.paths[commit_sha]

    return wrapper_memorize_paths



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['smartsquash']

package_data = \
{'': ['*']}

install_requires = \
['gitpython>=3.0.5,<4.0.0', 'loguru>=0.4.1,<0.5.0']

entry_points = \
{'console_scripts': ['sq = smartsquash.__main__:main']}

setup_kwargs = {
    'name': 'smartsquash',
    'version': '0.1.3',
    'description': 'Makes daily git workflows easier, automates rebases or fixups.',
    'long_description': "# smartsquash\n\nMakes daily git workflows easier, automates rebases or fixups.\n\n### build\n\n```sh\npoetry install --no-dev --no-root\npoetry build\n```\n\n### installation\n\n```sh\npip3 install dist/smartsquash-0.1.0-py3-none-any.whl\n```\n\n### usage\n\n```sh\nusage: sq [-h] [--target-branch TARGET_BRANCH] [--repo REPO] [--dry] [-s]\n\noptional arguments:\n  -h, --help            show this help message and exit\n  --target-branch TARGET_BRANCH\n                        Specify branch to target. Default is 'master'\n  --repo REPO           Specify repo to modify. Uses pwd by default\n  --dry                 Run dry\n  -s, --squash          Squash similar commits on your feature branch\n```\n\n### run tests\n\n```sh\npoetry run coverage run --source . -m pytest  \npoetry run coverage report\n```\n",
    'author': 'Max Wittig',
    'author_email': 'max.wittig95@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/max-wittig/smartsquash',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import argparse
import git
import sys
from pathlib import Path
from smartsquash import sq, helpers
from typing import Dict, Any
from loguru import logger


def get_args() -> dict:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--target-branch",
        type=str,
        required=False,
        default="master",
        help="Specify branch to target. Default is 'master'",
    )
    parser.add_argument(
        "--repo",
        type=str,
        required=False,
        default=Path.cwd().absolute(),
        help="Specify repo to modify. Uses pwd by default",
    )
    parser.add_argument(
        "--dry", required=False, default=False, action="store_true", help="Run dry"
    )
    parser.add_argument(
        "-s",
        "--squash",
        help="Squash similar commits on your feature branch",
        required=False,
        default=False,
        action="store_true",
    )
    return vars(parser.parse_args())


def setup_logger():
    logger.remove()
    logger.add(sys.stdout, colorize=True, format="<level>{message}</level>")


def main():
    setup_logger()
    args: Dict[str, Any] = get_args()
    repo_path: str = args.get("repo")
    target_branch: str = args.get("target_branch")
    dry: bool = args.get("dry")

    repo: git.Repo = helpers.get_repo(repo_path, target_branch)
    sq.fixup(target_branch, repo, dry)
    if args.get("squash"):
        sq.squash(target_branch, repo, dry)
